@@ -1,3 +1,10 @@
+"""
+B{ComputeDataService} Module: A central implementation of the L{ComputeDataService}
+
+A Meta-Scheduling service for pilots (both PilotCompute and PilotData)
+
+"""
+
 import sys
 import os
 import time
@@ -5,31 +12,32 @@ import threading
 import logging
 import pdb
 import Queue
-import saga
 import uuid
 import traceback
 import urlparse
 
-import bigjob
-import pilot
 from bigjob import logger
-from pilot.api import ComputeDataService, ComputeUnit, State
-from pilot.impl.pilotdata_manager import PilotData, DataUnit
+from pilot.api import ComputeDataService, State
+from pilot.impl.pilotdata_manager import DataUnit
+from pilot.impl.pilotcompute_manager import ComputeUnit
+
+
 #from pilot.coordination.advert import AdvertCoordinationAdaptor as CoordinationAdaptor
 from pilot.coordination.nocoord import NoCoordinationAdaptor as CoordinationAdaptor
+
+
 """ Loaded Module determines scheduler:
     
-    bigdata.scheduler.data_compute_scheduler - selects random locations for PD and WUs
+    bigdata.scheduler.data_compute_scheduler - selects random locations for PD and CUs
     bigdata.scheduler.data_compute_affinity_scheduler - considers affinity descriptions
     
 """
-
 from pilot.scheduler.data_compute_affinity_scheduler import Scheduler
 
 class ComputeDataService(ComputeDataService):
-    """ TROY WorkDataService.
+    """ B{ComputeDataService}
     
-        The WorkDataService is the application's interface to submit 
+        The ComputeDataService is the application's interface to submit 
         ComputeUnits and PilotData/DataUnit to the Pilot-Manager 
         in the P* Model.
     """    
@@ -37,10 +45,10 @@ class ComputeDataService(ComputeDataService):
 
 
     def __init__(self, cds_url=None):
-        """ Create a Work Data Service object.
+        """ Create a ComputeDataService object.
 
             Keyword arguments:
-            cds_url -- Reconnect to an existing WDS (optional).
+            cds_url -- Reconnect to an existing CDS (optional).
         """
         # Pilot Data
         self.data_units={}
@@ -53,8 +61,7 @@ class ComputeDataService(ComputeDataService):
         if cds_url == None:
             self.id=self.CDS_ID_PREFIX + str(uuid.uuid1())
             application_url = CoordinationAdaptor.get_base_url(pilot.application_id)
-            self.url = CoordinationAdaptor.add_cds(application_url, self)
-            
+            self.url = CoordinationAdaptor.add_cds(application_url, self)            
         else:
             self.id = self.__get_cds_id(cds_url)
             self.url = cds_url
@@ -65,7 +72,9 @@ class ComputeDataService(ComputeDataService):
         self.du_queue = Queue.Queue()
         self.stop=threading.Event()
         self.scheduler_thread=threading.Thread(target=self._scheduler_thread)
+        self.scheduler_thread.daemon=True
         self.scheduler_thread.start()
+        
 
     def __get_cds_id(self, cds_url):
         start = cds_url.index(self.CDS_ID_PREFIX)
@@ -74,45 +83,35 @@ class ComputeDataService(ComputeDataService):
 
 
     ###########################################################################
-    # Pilot Job
-    
-    def add_pilot_compute_service(self, pjs):
-        """ Add a PilotJobService to this CDS.
+    # Pilot Compute    
+    def add_pilot_compute_service(self, pcs):
+        """ Add a PilotComputeService to this CDS.
 
-            Keyword arguments:
-            pilotjob_services -- The PilotJob Service(s) to which this 
-                                 Work Unit Service will connect.
+            @param pcs: The PilotComputeService to which this ComputeDataService will connect.
 
-            Return:
-            Result
         """
-        self.pilot_job_services.append(pjs)
+        self.pilot_job_services.append(pcs)
         CoordinationAdaptor.update_cds(self.url, self)
+        
 
-    def remove_pilot_compute_service(self, pjs):
+    def remove_pilot_compute_service(self, pcs):
         """ Remove a PilotJobService from this CDS.
 
-            Note that it won't cancel the PilotJobService, it will just no
-            longer be connected to this WUS.
+            Note that it won't cancel the PilotComputeService, it will just no
+            longer be connected to this CDS.
 
             Keyword arguments:
-            pilotjob_services -- The PilotJob Service(s) to remove from this
-                                 Work Unit Service. 
-
-            Return:
-            Result
+            @param pcs: The PilotComputeService to remove from this ComputeDataService. 
         """
-        self.pilot_job_services.remove(pjs)
+        self.pilot_job_services.remove(pcs)
         CoordinationAdaptor.update_cds(self.url, self)
 
+
     def submit_compute_unit(self, compute_unit_description):
-        """ Submit a WU to this Work Unit Service.
+        """ Submit a CU to this Compute Data Service.
 
-            Keyword argument:
-            cud -- The ComputeUnitDescription from the application
-
-            Return:
-            ComputeUnit object
+            @param compute_unit_description: The ComputeUnitDescription from the application
+            @return: ComputeUnit object
         """
         cu = ComputeUnit(compute_unit_description, self)
         self.compute_units[cu.id]=cu
@@ -126,24 +125,14 @@ class ComputeDataService(ComputeDataService):
     def add_pilot_data_service(self, pds):
         """ Add a PilotDataService 
 
-            Keyword arguments:
-            pds -- The PilotDataService to add.
-
-            Return:
-            None
+            @param pds: The PilotDataService to add.
         """
         self.pilot_data_services.append(pds)
         CoordinationAdaptor.update_cds(self.url, self)
     
     def remove_pilot_data_service(self, pds):
-
         """ Remove a PilotDataService 
-            
-            Keyword arguments:
-            pds -- The PilotDataService to remove 
-            
-            Return:
-            None
+            @param pds: The PilotDataService to remove 
         """
         self.pilot_data_services.remove(pds)
         CoordinationAdaptor.update_cds(self.url, self)
@@ -183,60 +172,61 @@ class ComputeDataService(ComputeDataService):
     
     def cancel(self):
         """ Cancel the CDS. 
-            All associated PC objects are deleted.            
-            
-            Keyword arguments:
-            None
-
-            Return:
-            None
+            All associated PD and PC objects are canceled.            
         """
         # terminate background thread
         self.stop.set()
         CoordinationAdaptor.delete_cds(self.url)
    
     def wait(self):
-        """ Waits for CUs and DUs
-            Return if all du's are running 
-                   AND
-                   cu's are done
-            
+        """ Waits for CUs and DUs. Return after all DU's have been placed (i.e. in state Running)
+            and all CU's have been completed (i.e. in state Done) or if a fault has occurred or
+            the user has cancelled a CU or DU.            
         """
-        logger.debug("### START WAIT ###")
-        self.cu_queue.join()
-        logger.debug("CU queue empty")        
-        self.du_queue.join()
-        logger.debug("DU queue empty")        
-
-        for i in self.data_units.values():
-            i.wait()
-        logger.debug("DUs done")        
-            
-        for i in self.compute_units.values():
-            i.wait()     
-        logger.debug("CUs done")        
-               
-        logger.debug("### END WAIT ###")
-
-        #[pc.wait() for i in self.pilot_job_services for pc in i.list_pilots()]
-        #[pd.wait() for i in self.pilot_data_services for pd in i.list_pilots()]
+        try:
+            logger.debug("### START WAIT ###")
+            self.cu_queue.join()
+            logger.debug("CU queue empty")        
+            self.du_queue.join()
+            logger.debug("DU queue empty")        
+    
+            for i in self.data_units.values():
+                i.wait()
+            logger.debug("DUs done")        
+                
+            for i in self.compute_units.values():
+                i.wait()     
+            logger.debug("CUs done")        
+                   
+            logger.debug("### END WAIT ###")
+        except:
+            logger.debug("Ctrl-c detected. Terminating ComputeDataService...")
+            self.cancel()
+            raise KeyboardInterrupt
                 
         
     def get_state(self):
+        "@return: State of the ComputeDataService"
         return self.state
     
     
     def get_id(self):
+        "@return: id of ComputeDataService"
         return str(self.id)
+    
+    
+    def __del__(self):
+        """ Make sure that background thread terminates"""
+        self.cancel()
    
     ###########################################################################
     # Internal Scheduling
     def __update_scheduler_resources(self):
-        logging.debug("__update_scheduler_resources")        
+        logger.debug("__update_scheduler_resources")        
         pd = [s for i in self.pilot_data_services for s in i.list_pilots()]
         self.scheduler.set_pilot_data(pd)
         pj = [p for i in self.pilot_job_services for p in i.list_pilots()]
-        logging.debug("Pilot-Jobs: " + str(pj))
+        logger.debug("Pilot-Jobs: " + str(pj))
         self.scheduler.set_pilot_jobs(pj)
     
     def _schedule_du(self, du):
@@ -246,13 +236,13 @@ class ComputeDataService(ComputeDataService):
                 1.) Add all resources managed by the contained PDS 
                 2.) Select one resource
         """ 
-        logging.debug("Schedule PD")
+        logger.debug("Schedule PD")
         self.__update_scheduler_resources()
         selected_pilot_data = self.scheduler.schedule_pilot_data(du.data_unit_description)
         return selected_pilot_data 
     
     def _schedule_cu(self, cu):
-        logging.debug("Schedule CU")
+        logger.debug("Schedule CU")
         self.__update_scheduler_resources()
         selected_pilot_job = self.scheduler.schedule_pilot_job(cu.compute_unit_description)
         return selected_pilot_job
@@ -260,15 +250,15 @@ class ComputeDataService(ComputeDataService):
     def _scheduler_thread(self):
         while True and self.stop.isSet()==False:            
             try:
-                logging.debug("Scheduler Thread: " + str(self.__class__) + " Pilot Data")
+                logger.debug("Scheduler Thread: " + str(self.__class__) + " Pilot Data")
                 du = self.du_queue.get(True, 1)  
                 # check whether this is a real du object  
                 if isinstance(du, DataUnit):
                     pd=self._schedule_du(du)                
                     if(pd!=None):                        
-                        logging.debug("Initiate Transfer to PD.")
+                        logger.debug("Initiate Transfer to PD.")
                         du.add_pilot_data(pd)
-                        logging.debug("Transfer to PD finished.")
+                        logger.debug("Transfer to PD finished.")
                         du.update_state(State.Running) 
                         self.du_queue.task_done()                   
                     else:
@@ -278,17 +268,18 @@ class ComputeDataService(ComputeDataService):
                 pass
                     
             try:    
-                logging.debug("Scheduler Thread: " + str(self.__class__) + " Pilot Job")
+                logger.debug("Scheduler Thread: " + str(self.__class__) + " Pilot Job")
                 cu = self.cu_queue.get(True, 1)                
                 if isinstance(cu, ComputeUnit):                    
                     pj=self._schedule_cu(cu) 
                     if pj !=None:
                         cu = self.__expand_working_directory(cu, pj)                        
-                        pj.submit_cu(cu)           
+                        pj._submit_cu(cu)           
                         self.cu_queue.task_done()         
                     else:
+                        logger.debug("No resource found.")
                         self.cu_queue.task_done() 
-                        self.cu_queue.put(du)
+                        self.cu_queue.put(cu)
             except Queue.Empty:
                 pass
             except:
@@ -300,7 +291,7 @@ class ComputeDataService(ComputeDataService):
                               limit=2, file=sys.stderr)
             time.sleep(5)        
 
-        logging.debug("Re-Scheduler terminated")
+        logger.debug("Re-Scheduler terminated")
     
     
     def __expand_working_directory(self, compute_unit, pilot_job):
@@ -338,7 +329,7 @@ class ComputeDataService(ComputeDataService):
                     for du in pd.list_data_units():
                         logger.debug("DU URL:%s"%(du.url))
                         if du.url == pilot_data_url:
-                            logging.debug("Found PD %s at %s"%(du.url, pd.service_url))
+                            logger.debug("Found PD %s at %s"%(du.url, pd.service_url))
                             target_pd = pd 
                             target_du = du
                             break
@@ -350,7 +341,7 @@ class ComputeDataService(ComputeDataService):
                     components = urlparse.urlparse(pd_url)
                     compute_unit.compute_unit_description["working_directory"] = components.path
                     compute_unit._update_compute_unit_description(compute_unit.compute_unit_description)
-                    logging.debug("__expand_working_directory %s: Set working directory to %s"%(pilot_data_url, compute_unit.compute_unit_description["working_directory"]))
+                    logger.debug("__expand_working_directory %s: Set working directory to %s"%(pilot_data_url, compute_unit.compute_unit_description["working_directory"]))
                     return compute_unit
          
         return compute_unit
@@ -365,81 +356,17 @@ class ComputeDataService(ComputeDataService):
     def __find_pd_at_pj_resource(self, pilotjob):
         pass
    
-    
-    
-class ComputeUnit(ComputeUnit):
-    """ ComputeUnit - Wrapper for BigJob subjob """
-    CU_ID_PREFIX="cu-"  
 
-    def __init__(self, compute_unit_description, work_data_service=None):
-        self.id = self.CU_ID_PREFIX + str(uuid.uuid1())
-        if work_data_service!=None:
-            self.url = work_data_service.url + "/" + self.id
-            logger.debug("Created CU: %s"%self.url)  
-        self.state = State.New       
-        self.subjob = None # reference to BigJob Subjob 
-        self.compute_unit_description = compute_unit_description # CU Description
-        self.subjob_description = self.__translate_cu_sj_description(compute_unit_description)
-           
-                
-    def get_id(self):
-        return self.id
-    
-    
-    def get_state(self):
-        if self.subjob != None:
-            return self.subjob.get_state()
-        return self.state
-    
-    
-    def wait(self):
-        """ Wait until in Done state 
-            (or Failed state)
-        """
-        while True:
-            state = self.get_state()
-            logger.debug("Compute Unit - State: %s"%state)            
-            if state==State.Done or state==State.Failed:
-                break
-            time.sleep(2)
-            
-        #logger.debug("### END CU wait")
+###############################################################################
+# Unimplemented entities
 
-    
-    def cancel(self):
-        if self.subjob != None:
-            return self.subjob.cancel()
-        return None
-    
-    def __repr__(self):
-        return self.id
+class ComputeUnitService():
+    def __init__(self):
+        raise NotImplementedError("Please use ComputeDataService.")
 
+
+class DataUnitService():
+    def __init__(self):
+        raise NotImplementedError("Please use ComputeDataService.")
     
-    def _update_compute_unit_description(self, compute_unit_description):
-        self.compute_unit_description = compute_unit_description # WU Description
-        self.subjob_description = self.__translate_cu_sj_description(compute_unit_description)
-        
-    # INTERNAL
-    def __translate_cu_sj_description(self, compute_unit_description):
-        jd = saga.job.description()
-        if compute_unit_description.has_key("executable"): 
-            jd.executable = compute_unit_description["executable"]
-        jd.spmd_variation = "single"
-        if compute_unit_description.has_key("arguments"): 
-            jd.arguments = compute_unit_description["arguments"]
-        
-        if compute_unit_description.has_key("number_of_processes"):
-            jd.number_of_processes=str(compute_unit_description["number_of_processes"])
-        else:
-            jd.number_of_processes="1"
-        
-        if compute_unit_description.has_key("working_directory"): 
-            jd.working_directory = compute_unit_description["working_directory"]
-        if compute_unit_description.has_key("output"): 
-            jd.output =  compute_unit_description["output"]
-        if compute_unit_description.has_key("error"): 
-            jd.error = compute_unit_description["error"]
-        if compute_unit_description.has_key("file_transfer"):
-            jd.file_transfer=compute_unit_description["file_transfer"]            
-        return jd
-        
+
